/* UsiFe FileSystem
----------------
Copyright (2012): Rohan Sharma and Saurabh Kumar

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

CONTACT: 
Rohan Sharma : rickyrohan14@gmail.com
Saurabh Kumar : saurabh.catch@gmail.com
Smruti R. Sarangi : srsarangi@cse.iitd.ac.in
*/
#include "fuse_include.h" 
#include "do_create.h"

int do_check_split (const char *path,char **dirname,char **basename)
{
	char *cpath=strdup(path);
	char *tmp;
	tmp = strrchr(cpath, '/');
	if (tmp == NULL) {
		DEBUG_PRINTF("this should not happen %s\n", path);
		free(cpath);
		return -ENOENT;
	}
	*tmp='\0';
	tmp++;
	if (strlen(tmp) > 255) {
		DEBUG_PRINTF("basename exceeds 255 characters %s\n",path);
		free(cpath);
		return -ENAMETOOLONG;
	}
	*dirname=cpath;
	*basename=tmp;
	return 0;
}

int do_mkdir (const char *path, mode_t mode)
{
	int rt;
	time_t tm;
	errcode_t rc;

	char *p_path;
	char *r_path;

	ext2_ino_t ino;
	struct ext2_inode inode;

	struct fuse_context *ctx;

	ext2_filsys e2fs = (ext2_filsys)current_ext2fs();

	DEBUG_PRINTF("enter\n");
	DEBUG_PRINTF("path = %s, mode: 0%o, dir:0%o\n", path, mode, LINUX_S_IFDIR);

	rt=do_check_split(path, &p_path ,&r_path);
	if (rt != 0) {
		DEBUG_PRINTF("do_check(%s); failed\n", path);
		return rt;
	}

	DEBUG_PRINTF("parent: %s, child: %s, pathmax: %d\n", p_path, r_path, PATH_MAX);

	rt = do_readinode(e2fs, p_path, &ino, &inode);
	if (rt) {
		DEBUG_PRINTF("do_readinode(%s, &ino, &inode); failed\n", p_path);
		free(p_path);
		return rt;
	}

	do {
		DEBUG_PRINTF("calling ext2fs_mkdir(e2fs, %d, 0, %s);\n", ino, r_path);
		rc = ext2fs_mkdir(e2fs, ino, 0, r_path);
		if (rc == EXT2_ET_DIR_NO_SPACE) {
			DEBUG_PRINTF("calling ext2fs_expand_dir(e2fs, &d)\n", ino);
			if (ext2fs_expand_dir(e2fs, ino)) {
				DEBUG_PRINTF("error while expanding directory %s (%d)\n", p_path, ino);
				free(p_path);
				return -ENOSPC;
			}
		}
	} while (rc == EXT2_ET_DIR_NO_SPACE);
	if (rc) {
		DEBUG_PRINTF("ext2fs_mkdir(e2fs, %d, 0, %s); failed (%d)\n", ino, r_path, (int)rc);
		DEBUG_PRINTF("e2fs: %p, e2fs->inode_map: %p\n", e2fs, e2fs->inode_map);
		free(p_path);
		return -EIO;
	}

	rt = do_readinode(e2fs, path, &ino, &inode);
	if (rt) {
		DEBUG_PRINTF("do_readinode(%s, &ino, &inode); failed\n", path);
		return -EIO;
	}
	tm = e2fs->now ? e2fs->now : time(NULL);
	inode.i_mode = LINUX_S_IFDIR | mode;
	inode.i_ctime = inode.i_atime = inode.i_mtime = tm;
	ctx = fuse_get_context();
	if (ctx) {
		inode.i_uid = ctx->uid;
		inode.i_gid = ctx->gid;
	}
	rc = ext2fs_write_inode(e2fs, ino, &inode);
	if (rc) {
		DEBUG_PRINTF("ext2fs_write_inode(e2fs, ino, &inode); failed\n");
		free(p_path);
		return -EIO;
	}

	/* update parent dir */
	rt = do_readinode(e2fs, p_path, &ino, &inode);
	if (rt) {
		DEBUG_PRINTF("do_readinode(%s, &ino, &inode); dailed\n", p_path);
		free(p_path);
		return -EIO;
	}
	inode.i_ctime = inode.i_mtime = tm;
	rc = ext2fs_write_inode(e2fs, ino, &inode);
	if (rc) {
		DEBUG_PRINTF("ext2fs_write_inode(e2fs, ino, &inode); failed\n");
		free(p_path);
		return -EIO;
	}

	free(p_path);

	DEBUG_PRINTF("leave");
	return 0;
}

int do_modetoext2lag (mode_t mode)
{
	if (S_ISREG(mode)) {
		return EXT2_FT_REG_FILE;
	} else if (S_ISDIR(mode)) {
		return EXT2_FT_DIR;
	} else if (S_ISCHR(mode)) {
		return EXT2_FT_CHRDEV;
	} else if (S_ISBLK(mode)) {
		return EXT2_FT_BLKDEV;
	} else if (S_ISFIFO(mode)) {
		return EXT2_FT_FIFO;
	} else if (S_ISSOCK(mode)) {
		return EXT2_FT_SOCK;
	} else if (S_ISLNK(mode)) {
		return EXT2_FT_SYMLINK;
	}
	return EXT2_FT_UNKNOWN;
}

static inline int old_valid_dev(dev_t dev)
{
	return major(dev) < 256 && minor(dev) < 256;
}

static inline __u16 old_encode_dev(dev_t dev)
{
	return (major(dev) << 8) | minor(dev);
}

static inline __u32 new_encode_dev(dev_t dev)
{
	unsigned major = major(dev);
	unsigned minor = minor(dev);
	return (minor & 0xff) | (major << 8) | ((minor & ~0xff) << 12);
}

int op_create (ext2_filsys e2fs, const char *path, mode_t mode, dev_t dev, const char *fastsymlink)
{
	int rt;
	time_t tm;
	errcode_t rc;

	char *p_path;
	char *r_path;

	ext2_ino_t ino;
	struct ext2_inode inode;
	ext2_ino_t n_ino;

	struct fuse_context *ctx;

	DEBUG_PRINTF("enter\n");
	DEBUG_PRINTF("path = %s, mode: 0%o\n", path, mode);

	rt=do_check_split(path, &p_path, &r_path);

        if (rt != 0) {
		DEBUG_PRINTF("do_check split(%s); failed\n", path);
		return rt;
	}

	DEBUG_PRINTF("parent: %s, child: %s\n\n", p_path, r_path);

	rt = do_readinode(e2fs, p_path, &ino, &inode);
	if (rt) {
		DEBUG_PRINTF("do_readinode(%s, &ino, &inode); failed\n\n", p_path);
		free(p_path);
		return rt;
	}

	rc = ext2fs_new_inode(e2fs, ino, mode, 0, &n_ino);
	if (rc) {
		DEBUG_PRINTF("ext2fs_new_inode(ep.fs, ino, mode, 0, &n_ino); failed\n\n");
		return -ENOMEM;
	}

	do {
		DEBUG_PRINTF("calling ext2fs_link(e2fs, %d, %s, %d, %d);\n\n", ino, r_path, n_ino, do_modetoext2lag(mode));
		rc = ext2fs_link(e2fs, ino, r_path, n_ino, do_modetoext2lag(mode));
		if (rc == EXT2_ET_DIR_NO_SPACE) {
			DEBUG_PRINTF("calling ext2fs_expand_dir(e2fs, &d)\n\n", ino);
			if (ext2fs_expand_dir(e2fs, ino)) {
				DEBUG_PRINTF("error while expanding directory %s (%d)\n\n", p_path, ino);
				free(p_path);
				return -ENOSPC;
			}
		}
	} while (rc == EXT2_ET_DIR_NO_SPACE);
	if (rc) {
		DEBUG_PRINTF("ext2fs_link(e2fs, %d, %s, %d, %d); failed\n\n", ino, r_path, n_ino, do_modetoext2lag(mode));
		free(p_path);
		return -EIO;
	}

	if (ext2fs_test_inode_bitmap(e2fs->inode_map, n_ino)) {
		DEBUG_PRINTF("inode already set\n\n");
	}

	ext2fs_inode_alloc_stats2(e2fs, n_ino, +1, 0);
	memset(&inode, 0, sizeof(inode));
	tm = e2fs->now ? e2fs->now : time(NULL);
	inode.i_mode = mode;
	inode.i_atime = inode.i_ctime = inode.i_mtime = tm;
	inode.i_links_count = 1;
	inode.i_size = 0;
	ctx = fuse_get_context();
	if (ctx) {
		inode.i_uid = ctx->uid;
		inode.i_gid = ctx->gid;
	}

	/*if (S_ISCHR(mode) || S_ISBLK(mode)) {
		if (old_valid_dev(dev))
			inode.i_block[0]= ext2fs_cpu_to_le32(old_encode_dev(dev));
		else
			inode.i_block[1]= ext2fs_cpu_to_le32(new_encode_dev(dev));
	}*/

	if (S_ISLNK(mode) && fastsymlink != NULL) {
		inode.i_size = strlen(fastsymlink);
		strncpy((char *)&(inode.i_block[0]),fastsymlink,
				(EXT2_N_BLOCKS * sizeof(inode.i_block[0])));
	}

	rc = ext2fs_write_new_inode(e2fs, n_ino, &inode);
	if (rc) {
		DEBUG_PRINTF("ext2fs_write_new_inode(e2fs, n_ino, &inode);\n\n");
		free(p_path);
		return -EIO;
	}

	/* update parent dir */
	rt = do_readinode(e2fs, p_path, &ino, &inode);
	if (rt) {
		DEBUG_PRINTF("do_readinode(%s, &ino, &inode); dailed\n\n", p_path);
		free(p_path);
		return -EIO;
	}
	inode.i_ctime = inode.i_mtime = tm;
	rc = ext2fs_write_inode(e2fs, ino, &inode);
	if (rc) {
		DEBUG_PRINTF("ext2fs_write_inode(e2fs, ino, &inode); failed\n\n");
		free(p_path);
		return -EIO;
	}

	free(p_path);
      
        fflush(stdout);
	DEBUG_PRINTF("leave\n");
	return 0;
}

int do_create (const char *path, mode_t mode, struct fuse_file_info *fi)
{
	int rt;
	ext2_filsys e2fs = (ext2_filsys)current_ext2fs();

	DEBUG_PRINTF("enter\n");
	DEBUG_PRINTF("path = %s, mode: 0%o\n\n", path, mode);

	if (do_open(path, fi) == 0) {
		DEBUG_PRINTF("leave file exist\n\n");
		return 0;
	}

	rt = op_create(e2fs, path, mode, 0, NULL);
	if (rt != 0) {
		return rt;
	}

	if (do_open(path, fi)) {
		DEBUG_PRINTF("op_open(path, fi); failed\n\n");
		return -EIO;
	}

	DEBUG_PRINTF("leave\n\n");
	return 0;
}


