/* UsiFe FileSystem
----------------
Copyright (2012): Rohan Sharma and Saurabh Kumar

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

CONTACT: 
Rohan Sharma : rickyrohan14@gmail.com
Saurabh Kumar : saurabh.catch@gmail.com
Smruti R. Sarangi : srsarangi@cse.iitd.ac.in
*/
#include "fuse_include.h" 
#include "do_delete.h"


void free_split (char *dirname, char *basename)
{
	free(dirname);
}

int do_unlink (const char *path)
{
	int rt;
	errcode_t rc;

	char *p_path;
	char *r_path;

	ext2_ino_t p_ino;
	struct ext2_inode p_inode;
	ext2_ino_t r_ino;
	//struct ext2_vnode *r_vnode;
	struct ext2_inode r_inode;

	ext2_filsys e2fs = (ext2_filsys)current_ext2fs();

	DEBUG_PRINTF("enter");
	DEBUG_PRINTF("path = %s", path);

	rt=do_check_split(path, &p_path, &r_path);
	if (rt != 0) {
		DEBUG_PRINTF("do_check_split: failed");
		return rt;
	}

	DEBUG_PRINTF("parent: %s, child: %s", p_path, r_path);

	rt = do_readinode(e2fs, p_path, &p_ino, &p_inode);
	if (rt) {
		DEBUG_PRINTF("do_readinode(%s, &p_ino, &p_inode); failed", p_path);
		free_split(p_path, r_path);
		return rt;
	}
	rt = do_readinode(e2fs, path, &r_ino, &r_inode);
	if (rt) {
		DEBUG_PRINTF("do_readvnode(%s, &r_ino, &r_vnode); failed", path);
		free_split(p_path, r_path);
		return rt;

	}
	//r_inode = vnode2inode(r_vnode);

	if(LINUX_S_ISDIR(r_inode.i_mode)) {
		DEBUG_PRINTF("%s is a directory", path);
		//vnode_put(r_vnode,0);
		free_split(p_path, r_path);
		return -EISDIR;
	}

	rc = ext2fs_unlink(e2fs, p_ino, r_path, r_ino, 0);
	if (rc) {
		DEBUG_PRINTF("ext2fs_unlink(e2fs, %d, %s, %d, 0); failed", p_ino, r_path, r_ino);
		//vnode_put(r_vnode,0);
		free_split(p_path, r_path);
		return -EIO;
	}

	if (r_inode.i_links_count > 0) {
		r_inode.i_links_count -= 1;
	}

	p_inode.i_ctime = p_inode.i_mtime = e2fs->now ? e2fs->now : time(NULL);
         rc = ext2fs_write_inode(e2fs, p_ino, &p_inode); 
	if (rc) {
		DEBUG_PRINTF("ext2fs_write_inode(e2fs, p_ino, &p_inode); failed");
		//vnode_put(r_vnode,1);
		free_split(p_path, r_path);
		return -EIO;
	}

	r_inode.i_ctime = e2fs->now ? e2fs->now : time(NULL);
	rc = ext2fs_write_inode(e2fs, r_ino, &r_inode); 
	if (rc) {
		DEBUG_PRINTF("vnode_put(r_vnode,1); failed");
		free_split(p_path, r_path);
		return -EIO;
	}

	free_split(p_path, r_path);
	DEBUG_PRINTF("leave");
	return 0;
}

struct rmdir_st {
	ext2_ino_t parent;
	int empty;
};


static int release_blocks_proc (ext2_filsys fs, blk_t *blocknr, int blockcnt, void *private)
{
	blk_t block;

	DEBUG_PRINTF("enter");

	block = *blocknr;
	ext2fs_block_alloc_stats(fs, block, -1);

	DEBUG_PRINTF("leave");
#ifdef CLEAN_UNUSED_BLOCKS
	*blocknr = 0;
	return BLOCK_CHANGED;
#else
	return 0;
#endif
}

int do_killfilebyinode (ext2_filsys e2fs, ext2_ino_t ino, struct ext2_inode *inode)
{
	errcode_t rc;
	char scratchbuf[3*e2fs->blocksize];

	DEBUG_PRINTF("enter");

	inode->i_links_count = 0;
	inode->i_dtime = time(NULL);

	rc = ext2fs_write_inode(e2fs, ino, inode);
	if (rc) {
		DEBUG_PRINTF("ext2fs_write_inode(e2fs, ino, inode); failed");
		return -EIO;
	}

	if (ext2fs_inode_has_valid_blocks(inode)) {
		DEBUG_PRINTF("start block delete for %d", ino);
#ifdef CLEAN_UNUSED_BLOCKS
		ext2fs_block_iterate(e2fs, ino, BLOCK_FLAG_DEPTH_TRAVERSE, scratchbuf, release_blocks_proc, NULL);
#else
		ext2fs_block_iterate(e2fs, ino, 0, scratchbuf, release_blocks_proc, NULL);
#endif
	}

	ext2fs_inode_alloc_stats2(e2fs, ino, -1, LINUX_S_ISDIR(inode->i_mode));

	DEBUG_PRINTF("leave");
	return 0;
}
static int rmdir_proc (ext2_ino_t dir EXT2FS_ATTR((unused)),
		       int entry EXT2FS_ATTR((unused)),
		       struct ext2_dir_entry *dirent,
		       int offset EXT2FS_ATTR((unused)),
		       int blocksize EXT2FS_ATTR((unused)),
		       char *buf EXT2FS_ATTR((unused)), void *private)
{
	int *p_empty= (int *) private;

	DEBUG_PRINTF("enter");
	DEBUG_PRINTF("walking on: %s", dirent->name);

	if (dirent->inode == 0 ||
			(((dirent->name_len & 0xFF) == 1) && (dirent->name[0] == '.')) ||
			(((dirent->name_len & 0xFF) == 2) && (dirent->name[0] == '.') && 
			 (dirent->name[1] == '.'))) {
		DEBUG_PRINTF("leave");
		return 0;
	}
	*p_empty = 0;
	DEBUG_PRINTF("leave (not empty)");
	return 0;
}

int do_check_empty_dir(ext2_filsys e2fs, ext2_ino_t ino)
{
	errcode_t rc;
	int empty = 1;

	rc = ext2fs_dir_iterate2(e2fs, ino, 0, 0, rmdir_proc, &empty);
	if (rc) {
		DEBUG_PRINTF("while iterating over directory");
		return -EIO;
	}

	if (empty == 0) {
		DEBUG_PRINTF("directory not empty");
		return -ENOTEMPTY;
	}

	return 0;
}

int do_rmdir (const char *path)
{
	int rt;
	errcode_t rc;

	char *p_path;
	char *r_path;

	ext2_ino_t p_ino;
	struct ext2_inode p_inode;
	ext2_ino_t r_ino;
	struct ext2_inode r_inode;
	
	ext2_filsys e2fs = (ext2_filsys)current_ext2fs();

	DEBUG_PRINTF("enter");
	DEBUG_PRINTF("path = %s", path);

	rt=do_check_split(path, &p_path, &r_path);
	if (rt != 0) {
		DEBUG_PRINTF("do_check_split: failed");
		return rt;
	}

	DEBUG_PRINTF("parent: %s, child: %s", p_path, r_path);
	
	rt = do_readinode(e2fs, p_path, &p_ino, &p_inode);
	if (rt) {
		DEBUG_PRINTF("do_readinode(%s, &p_ino, &p_inode); failed", p_path);
		free_split(p_path, r_path);
		return rt;
	}
	rt = do_readinode(e2fs, path, &r_ino, &r_inode);
	if (rt) {
		DEBUG_PRINTF("do_readinode(%s, &r_ino, &r_inode); failed", path);
		free_split(p_path, r_path);
		return rt;
		
	}
	if (!LINUX_S_ISDIR(r_inode.i_mode)) {
		DEBUG_PRINTF("%s is not a directory", path);
		free_split(p_path, r_path);
		return -ENOTDIR;
	}
	if (r_ino == EXT2_ROOT_INO) {
		DEBUG_PRINTF("root dir cannot be removed", path);
		free_split(p_path, r_path);
		return -EIO;
	}
	
	rt = do_check_empty_dir(e2fs, r_ino);
	if (rt) {
		DEBUG_PRINTF("do_check_empty_dir filed");
		free_split(p_path, r_path);
		return rt;
	}

	rc = ext2fs_unlink(e2fs, p_ino, r_path, r_ino, 0);
	if (rc) {
		DEBUG_PRINTF("while unlinking ino %d", (int) r_ino);
		free_split(p_path, r_path);
		return -EIO;
	}

	rt = do_killfilebyinode(e2fs, r_ino, &r_inode);
	if (rt) {
		DEBUG_PRINTF("do_killfilebyinode(r_ino, &r_inode); failed");
		free_split(p_path, r_path);
		return rt;
	}

	rt = do_readinode(e2fs, p_path, &p_ino, &p_inode);
	if (rt) {
		DEBUG_PRINTF("do_readinode(p_path, &p_ino, &p_inode); failed");
		free_split(p_path, r_path);
		return rt;
	}
	if (p_inode.i_links_count > 1) {
		p_inode.i_links_count--;
	}
	rc = ext2fs_write_inode(e2fs, p_ino, &p_inode);
	if (rc) {
		DEBUG_PRINTF("ext2fs_write_inode(e2fs, ino, inode); failed");
		free_split(p_path, r_path);
		return -EIO;
	}

	free_split(p_path, r_path);

	DEBUG_PRINTF("leave");
	return 0;
}
