/* UsiFe FileSystem
----------------
Copyright (2012): Rohan Sharma and Saurabh Kumar

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

CONTACT: 
Rohan Sharma : rickyrohan14@gmail.com
Saurabh Kumar : saurabh.catch@gmail.com
Smruti R. Sarangi : srsarangi@cse.iitd.ac.in
*/
#include "fuse_include.h" 
#include "do_read.h"



/*  
 * Read mode 4, described below.
 */
int do_read4 (const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
{
    __u64 pos;
    errcode_t rc;
    unsigned int bytes=0;
    ext2_file_t efile = (ext2_file_t)((unsigned long)(fi->fh));
    ext2_ino_t ino;
    ext2_filsys e2fs = (ext2_filsys)current_ext2fs();
    struct ext2_inode inode;

    DEBUG_PRINTF("enter  MODE 4 READ %d %d %d %d\n\n",(int)offset,(int)size,fi,efile);
    DEBUG_PRINTF("path = %s\n\n", path);

    memcpy(&inode,efile->inode,sizeof(struct ext2_inode));

    if(inode.i_file_acl!=0)                         
         {
          DEBUG_PRINTF("Reading encrypted blocks\n");           

	  unsigned int enc_blk_no=offset/4096; //encrypted block no.
          short index=0;
          int i=0;         
          int start_blk=0;
          int off=0;
          char data[4096];

          memcpy(&index,data,sizeof(short));  

          start_blk=inode.i_file_acl;

	  for(i=0;i<enc_blk_no;i++)
	   {
	    pthread_mutex_lock(&mutex);
            io_channel_read_blk(e2fs->io,start_blk,1, data);
            pthread_mutex_unlock(&mutex);
 
            memcpy(&index,data,sizeof(short)); 
            if(index<=510)
               goto exit;
	    start_blk=*((int *)(data+4092)); 

	   }
	    pthread_mutex_lock(&mutex);
            io_channel_read_blk(e2fs->io,start_blk,1, data);
            pthread_mutex_unlock(&mutex);

           off=offset-4096*enc_blk_no;
	
           bytes= (4096-off) < size ? (4096-off) : size;

           memcpy(buf,data+off,bytes);  
           DEBUG_PRINTF("Read enc blocks DONE %d\n",bytes); 
          }
	DEBUG_PRINTF("leave");
exit:
        fflush(stdout);
	return bytes;
}


/*  
 * Read mode 3, described below.
 */
int do_read3 (const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
{
    __u64 pos;
    errcode_t rc;
    unsigned int bytes;
    ext2_file_t efile = (ext2_file_t)((unsigned long)(fi->fh));
    ext2_ino_t ino;
    ext2_filsys e2fs = (ext2_filsys)current_ext2fs();
    struct ext2_inode inode;

    DEBUG_PRINTF("enter  MODE 3 RAED %d %d %d %d\n\n",(int)offset,(int)size,fi,efile);
    DEBUG_PRINTF("path = %s\n\n", path);

    memcpy(&inode,efile->inode,sizeof(struct ext2_inode));
    pthread_mutex_lock(&mutex);
    
    rc = ext2fs_file_llseek(efile, offset, SEEK_SET, &pos);
	if (rc) {
		 pthread_mutex_unlock(&mutex); 
                 return -EINVAL;
	}
    rc = ext2fs_file_read(efile, buf, size, &bytes);
	if (rc) {
		 pthread_mutex_unlock(&mutex);
                 return -EIO;
	}

    pthread_mutex_unlock(&mutex);

    if(inode.i_file_acl!=0)                         
         {
          DEBUG_PRINTF("Starting Decryption\n\n");           

          char data[4096];
          memset(data,0,4096);
          int block_start=offset/4096+1;
          int block_end=((offset+bytes-1)/4096)+1;
          short index=0;
          int *p;
          int i=0;
          char *out;
 
          short *start,*end; 

          DEBUG_PRINTF("Start %d end %d \n\n",block_start,block_end);

          pthread_mutex_lock(&mutex);
          io_channel_read_blk(e2fs->io,inode.i_file_acl,1, data);
          pthread_mutex_unlock(&mutex);

          memcpy(&index,data,sizeof(short));  
         
          int actual_start,actual_end;
          for(i=0;i<index;)
	   {
             
             if(index)
               {
                 if(index==511)
                  p=(int *)(data+2+509*8);
                 else                
                  p=(int *)(data+2+(index-1)*8);

                if((*p)<block_start)
                 {     

                      DEBUG_PRINTF("skipping block\n\n");
 
                      if(index<=510)
                         {
                                  DEBUG_PRINTF("exiting\n");
                                  goto exit; 
                             }
                      else if(index==511)
                         {     

                               DEBUG_PRINTF("Reloading block\n");
                               p=(int *)(data+4092); 
                              
                               pthread_mutex_lock(&mutex);
                               io_channel_read_blk(e2fs->io,*p,1, data);
                               pthread_mutex_unlock(&mutex);

                               memcpy(&index,data,sizeof(short));  
                               i=0;
                             } 
                    }
               }

             p=(int *)(data+2+i*8);
             start=(short *)(data+2+i*8+4);
             end=(short *)(data+2+i*8+6);
             DEBUG_PRINTF("Earliest break %d %d %d %d \n\n",*p,*start,*end,bytes);             

             if((*p)>=block_start)
             {  
               
               if((*p)>block_end)
                 break;        
               
               actual_start=((*p)-1)*4096+*start;
               actual_end=((*p)-1)*4096+*end;
               DEBUG_PRINTF("Earliest break1 %d %d %d\n",*p,actual_start,actual_end);              

               if(actual_start>offset+bytes || actual_end<=offset)
                 break;

               if(actual_start<=offset)
                        actual_start=offset+1;
               
               if(actual_end>offset+size)
                        actual_end=offset+bytes;

               DEBUG_PRINTF("Decryption SIZE %d %d %d BLOCK %d %d\n",actual_end-actual_start+1,actual_start,actual_end,i,index);

               simple_encryption(buf+actual_start-(offset+1),1,actual_end-actual_start+1,buf+actual_start-(offset+1));
             } 
            
              i++;
              if(i==510&&index==511)            
              {

                DEBUG_PRINTF("loading new ACL BLOCK do_read3\n");
                p=(int *)(data+4092); 
                pthread_mutex_lock(&mutex);
                io_channel_read_blk(e2fs->io,*p,1, data);
                pthread_mutex_unlock(&mutex);
                memcpy(&index,data,sizeof(short));  
                i=0;                
              }             
             }
            DEBUG_PRINTF("Decryption DONE\n\n"); 
          }
	DEBUG_PRINTF("leave");
        exit:
        	fflush(stdout);
		return bytes;
}


/*  
 * Read mode 2.
 */
int do_read2 (const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
{
   __u64 pos;
   errcode_t rc;
   unsigned int bytes;
   ext2_file_t efile = (ext2_file_t)((unsigned long)(fi->fh));

   pthread_mutex_lock(&mutex);
   rc = ext2fs_file_llseek(efile, offset, SEEK_SET, &pos);
   if (rc) {
		pthread_mutex_unlock(&mutex);
                return -EINVAL;
	   }

   rc = ext2fs_file_read(efile, buf, size, &bytes);
   if (rc) {
		pthread_mutex_unlock(&mutex);
                return -EIO;
	   }

   pthread_mutex_unlock(&mutex);
	
   int size1;
   size1=size;
   DEBUG_PRINTF("READ2 %d %d %d %d\n",offset,size1,size,bytes);
   fflush(stdout);
   return bytes;
}

/*
 * This function reads data from file.
 * There are three modes of reading.
 * MODE 1:This mode considers encrypted data as if it is not present in the file.It basically skips encrypted portion 
 *        of file and outputs only plain data in the file. 
 *
 * MODE 2:This mode reads data without decrypting data first or skipping through encrypted data.In this mode 
 *        complete file data is visible.  
 *
 * MODE 3:This mode decrypts encrypted data on the fly and outputs decrypted data to user.
 *
 *
 * 
 *
 */

int do_read (const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
 {
   /*  
    * Check for reading mode.
    */
   if(!gtk_override)
   {
    PWord_t  PValue; 
    int val;
    ext2_file_t efile = (ext2_file_t)((unsigned long)(fi->fh));
    JLG(PValue,PJLArray,efile->ino);
    if(PValue!=PJERR && PValue!=NULL)
      {
	DEBUG_PRINTF("PValue is %d\n",*PValue);
	val=((*PValue)&READ_MODE)>>1;
	DEBUG_PRINTF("Read Mode is %d\n",val);
        if(val==READ_MODE2)
    	 return do_read2(path,buf,size,offset,fi);
	else if(val==READ_MODE3)
     	 return do_read3(path,buf,size,offset,fi);
	//else if(val==READ_MODE4)
     	 //return do_read4(path,buf,size,offset,fi);
      }
    else if(PValue==NULL)	
      {
	DEBUG_PRINTF("PValue is null\n");
      }
    else
      {
	DEBUG_PRINTF("Hashing error\n");
        return -1;
      }		
   }
   else
   {
   if(mode==2)
     return do_read2(path,buf,size,offset,fi);
   else if(mode==3)
     return do_read3(path,buf,size,offset,fi);
   }


   ext2_ino_t ino;
   struct ext2_inode inode;
   ext2_filsys e2fs = (ext2_filsys)current_ext2fs();
   __u64 pos;
   errcode_t rc;
   ext2_file_t efile = (ext2_file_t)((unsigned long)(fi->fh));
   char buf1=2;

   DEBUG_PRINTF("READING %s\n",path);
   DEBUG_PRINTF("SIZE %d OFFSET %d\n",size,(int)offset);
   memcpy(&inode,efile->inode,sizeof(struct ext2_inode));

   char data[4096];
   int i=0,k=0;
   short range;
   int block;
   int j=0;
   int file_offset=0,p_file_offset=0,actual_offset=0,p_actual_offset=0;
   int dblock;
   int e_bytes,p_e_bytes;
   short start,end,p_start,p_end;
   int p;
     
   start=1;end=0;
   p_start=1;p_end=0;
   i=0;

   /*  
    * Check if file contains encrypted regions.
    */
   if(inode.i_file_acl!=0) 
     {
         pthread_mutex_lock(&mutex);
         io_channel_read_blk(e2fs->io,inode.i_file_acl,1, data);
         pthread_mutex_unlock(&mutex);

     }
   else
     {
      memset(data,0,4096);
     }

    short index; 
    int nodata=0;
    int bytes,a_bytes=0,read=0;
    int a_size=size;
    int pp_start=1,pp_dblock=0;

    memcpy(&index,data,sizeof(short));   
    memcpy(&e_bytes,data+2+510*8,sizeof(int));
    
    if(index<=510 && index!=0)
    { memcpy(&dblock,data+2+(index-1)*8,sizeof(int));
      memcpy(&start,data+2+(index-1)*8+4,sizeof(short));
      memcpy(&end,data+2+(index-1)*8+6,sizeof(short)); 
    }  
    else if(index==511 && index!=0)
    { memcpy(&dblock,data+2+(510-1)*8,sizeof(int));
      memcpy(&start,data+2+(510-1)*8+4,sizeof(short));
      memcpy(&end,data+2+(510-1)*8+6,sizeof(short)); 
    }  

   /*  
    * Since we are skipping some parts , we need to find actual offset within file for given logical file offset.
    */
    while(1)
    {
       DEBUG_PRINTF("Entered \n"); 
       p_actual_offset=actual_offset;
       p_file_offset=file_offset; 

       if(index==0)
         {
           DEBUG_PRINTF("NO DATA\n"); 
           nodata=1;
              break;
         }
         
       file_offset=(dblock-1)*4096+end-e_bytes;                 
       actual_offset=(dblock-1)*4096+end;

       if(file_offset>offset)
          {
            DEBUG_PRINTF("Done\n");
            file_offset=p_file_offset;
            actual_offset=p_actual_offset;
            p_start=pp_start;
            i=pp_dblock;
            p_actual_offset=actual_offset;
            p_file_offset=file_offset; 
            break;
          }
       else
          {
             if(index<511)
               {
                  DEBUG_PRINTF("NO DATA\n"); 
                  index=0;
                  nodata=1;
                  p_actual_offset=actual_offset;
                  p_file_offset=file_offset; 
                  break;        
                  }
             else if(index==511)
               {
                 memcpy(&p,data+4092,sizeof(int));
                 pthread_mutex_lock(&mutex); 
                 io_channel_read_blk(e2fs->io,p,1, data);
                 pthread_mutex_unlock(&mutex); 

                 pp_start=end+1;
                 pp_dblock=dblock-1; 

                 memcpy(&index,data,sizeof(short));   
                 memcpy(&p_e_bytes,data+2+510*8,sizeof(int));
    		 if(index<=510 && index!=0)
                 { memcpy(&dblock,data+2+(index-1)*8,sizeof(int));
                   memcpy(&start,data+2+(index-1)*8+4,sizeof(short));
                   memcpy(&end,data+2+(index-1)*8+6,sizeof(short)); }  
                 else if(index==511 && index!=0)
                 { memcpy(&dblock,data+2+(510-1)*8,sizeof(int));
                   memcpy(&start,data+2+(510-1)*8+4,sizeof(short));
                   memcpy(&end,data+2+(510-1)*8+6,sizeof(short)); }  

                 e_bytes=e_bytes+p_e_bytes;
               }
           }  
    	}

	DEBUG_PRINTF("FILE %d  ACTUAL %d \n",file_offset,actual_offset);

	pp_dblock++;
        /*  
         * Read contents of file.
         */
	while(1){
	   
	  if(index==511 && j==510)
             {
		    DEBUG_PRINTF("Reloading required\n");
		    j=0;
		    memcpy(&block,data+4092,sizeof(int));
		    pthread_mutex_lock(&mutex);
		    io_channel_read_blk(e2fs->io,block,1, data);

		    pthread_mutex_unlock(&mutex);
		    memcpy(&index,data,sizeof(short));
             }
	  else if((index==0)||(j==index && index!=511))
             {
		    DEBUG_PRINTF("NO more data available\n");
		    nodata=1; 
             }

	   if(nodata==1)
	    {
	      file_offset=offset;
	      actual_offset=actual_offset+file_offset-p_file_offset;
	      bytes=size;
	      goto read; 
	 
	     }
	    else if(nodata==0)
	    {
	      memcpy(&dblock,data+2+(j)*8,sizeof(int));
	      memcpy(&start,data+2+(j)*8+4,sizeof(short));
	      memcpy(&end,data+2+(j)*8+6,sizeof(short));
	      file_offset=file_offset+(dblock-pp_dblock)*4096+start-p_start;
	      actual_offset=actual_offset+(dblock-pp_dblock)*4096+end-p_start+1;
	      j++;  
	      p_start=end+1; 
	      pp_dblock=dblock;

	     }

	     if(file_offset>offset)
		     {
		        bytes=file_offset-offset;
		        file_offset=offset;
		        actual_offset=p_actual_offset+file_offset-p_file_offset;
		        
		        goto read;

		      }

		    back:
		    p_actual_offset=actual_offset;
		    p_file_offset=file_offset;

	}	

read:
      DEBUG_PRINTF("ACTUAL %d FILE %d BYTES %d\n",actual_offset,file_offset,bytes);
      pthread_mutex_lock(&mutex); 
      rc = ext2fs_file_llseek(efile, actual_offset  , SEEK_SET, &pos);
      if (rc) {
		printf("READING FAILED  %s \n",path);
                return -EINVAL;
                }
      read=bytes;
      if((bytes+a_bytes)>a_size)
               read=a_size-a_bytes;
      
      rc = ext2fs_file_read(efile, buf+a_bytes, read, &bytes);
      pthread_mutex_unlock(&mutex); 
       
      if(bytes==0)
          {
            DEBUG_PRINTF("EOF reached\n");
            goto done;
          }
      if (rc) {
          printf("READING FAILED  %s \n",path);	
          return -EIO;
	
             }
  
      DEBUG_PRINTF("SUCCESS READING %d \n",bytes);

   
      size=size-bytes;
      offset=offset+bytes;
      file_offset=file_offset+bytes;
      if(nodata==0)
	      actual_offset=actual_offset+bytes+end-start+1;
      else
	      actual_offset=actual_offset+bytes;	

      a_bytes+=bytes;

      if(size>0)
       {
         goto back;
       }

      done:
  	return a_bytes;

}

/*  
 * Old reading fuunction kept for debugging purpose.
 */
int do_add_extended_data(struct region *blocks,int i_file_acl,int b)
  {
      ext2_filsys e2fs = (ext2_filsys)current_ext2fs();
      char data[4096];
      memset(data,0,4096); 
      struct region1{
         int block;
         short start;
         short end; 
       };  
      struct region1 abcd;
      short index=0;
         
      if(i_file_acl){
         int  rc= io_channel_read_blk(e2fs->io,i_file_acl,1, data);
           
         memcpy(&index,data,sizeof(short));  
         int k=0,*p;
         char *temp;
         struct region1 *temp1;  
         struct region *previous,*temp2;
         for(k=1;k<=index;)
           {
              temp=data+(k-1)*sizeof(struct region1)+sizeof(short);
              temp1=(struct region1 *)temp;
              temp2=malloc(sizeof(struct region));
              temp2->next=NULL; 
              temp2->start=temp1->start;   
              temp2->end=temp1->end;
              temp2->set=1;
              
              if(temp1->block>(b+1)*10)
                  break;

              if(temp1->block>(b*10))
              {
                 temp1->block=((temp1->block-1)%10)+1;  
              DEBUG_PRINTF("Block data  %d\n",index);
              DEBUG_PRINTF("Block data  %d\n",temp1->block);
              DEBUG_PRINTF("Block data  %d\n",temp2->start);
              DEBUG_PRINTF("Block data  %d\n\n\n",temp2->end);                       


              if(!blocks[temp1->block-1].set)
                    { 
		      DEBUG_PRINTF("Iff...........%d\n",(int)temp2); 
                      (blocks[temp1->block-1])=*temp2;
                      previous=&blocks[temp1->block-1];
                      }
              else  { 
                    previous->next=temp2; 
                    previous=temp2;}
               }
               k++;  
               if(k==512)                        //if new block is to be loaded
                {
                 DEBUG_PRINTF("loading new ACL BLOCK\n\n");
                 p=(int *)(data+4092); 
                 io_channel_read_blk(e2fs->io,*p,1, data);
                 k=1;
                 memcpy(&index,data,sizeof(short));  
                 }      
             }  
          
          for(k=0;k<=1;k++)
            {
               previous=&blocks[k];  
               while(previous!=NULL&&(blocks[k].set!=0))
                    {
                      DEBUG_PRINTF("printing %d %d %d\n\n",k,previous->start,previous->end);
                      previous=previous->next;
                    }

             }
             fflush(stdout);
	}
       }


/*  
 * Old reading fuunction kept for debugging purpose.
 */
int do_read_old (const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
 {
   if(mode==2)
     return do_read2(path,buf,size,offset,fi);
   else if(mode==3)
     return do_read3(path,buf,size,offset,fi);

   ext2_ino_t ino;
   struct ext2_inode inode;
   ext2_filsys e2fs = (ext2_filsys)current_ext2fs();
   int b=0;
   struct region blocks[10]={0,0,0,0,0,0,0,0,0,0};
   char bitmap[8]={1,1,1,1,1,1,1,1}; //bitmap is not required
   __u64 pos;
   errcode_t rc;
   unsigned int bytes,bytes1;
   ext2_file_t efile = (ext2_file_t)((unsigned long)(fi->fh));
   char buf1=2;
   
   DEBUG_PRINTF("READING %s\n",path);
   DEBUG_PRINTF("SIZE %d  %d\n",size,(int)offset);
      
   bytes=bytes1=0; 

   int rt = do_readinode(e2fs,path, &ino, &inode);
   if(rt)
      { 
	DEBUG_PRINTF("failed to open inode\n\n\n");
        return -ENOENT; 
       }

   do_add_extended_data(blocks,inode.i_file_acl,b);
   b++;

   DEBUG_PRINTF("OFFSET %d\n",(int)offset);  
   fflush(stdout);   

   int k,i,j,file_offset=0,file_offset1=0,actual_offset=0,read_offset=0;
   int l=0;
   int read;
   struct region *abcd;
   
   k=0; 
   i=0;
   j=0;
   file_offset=file_offset1=actual_offset=read_offset=0;
     
   for(j=0;;j++)
     {
         DEBUG_PRINTF("Entered\n\n");
         int previous_end=0;
         
         if(j>=(b)*10)
           {
               DEBUG_PRINTF("do_read adding more data %d\n",b*10); 
               for(k=0;k<10;k++)
                   {blocks[k].set=0;
                    blocks[k].start=0;
                    blocks[k].end=0;}
               do_add_extended_data(blocks,inode.i_file_acl,b);
               b++;

           }
           abcd=&blocks[(j)%10];         

         while(1){

           DEBUG_PRINTF("Entered 2\n\n");

           if(abcd==NULL||abcd->set==0)
             { DEBUG_PRINTF("Entered 3\n\n");
               file_offset1+=4096-previous_end;     
               if(file_offset1>offset)
               {
                  actual_offset=(j)*4096+previous_end+offset-file_offset;
                  read_offset= (j+1)*4096;

                  goto out; 
                  back:
	                  file_offset=file_offset1;
               
                 }
                file_offset=file_offset1;
                break;
               
             }
          
           DEBUG_PRINTF("Entered 4\n\n");
           file_offset1+=abcd->start-1-previous_end;
           if(file_offset1>offset)
              {
                actual_offset=(j)*4096+previous_end+offset-file_offset; 
                read_offset=(j)*4096+abcd->start-1 ; 
                goto out1;  
                back1:
                	file_offset=file_offset1; 
             }
           file_offset=file_offset1;
           previous_end=abcd->end; 
                      
           abcd=(abcd->next); 
          }
     }
       
out:
       DEBUG_PRINTF("ACTUAL OFFSET %d \n",actual_offset);
       DEBUG_PRINTF("Reading till  %d \n",read_offset); 
   
       rc = ext2fs_file_llseek(efile,actual_offset  , SEEK_SET, &pos);
       if (rc) {
		DEBUG_PRINTF("READING FAILED %s \n",path);
                return -EINVAL;
                }
       read=read_offset-actual_offset;
       if((bytes+read_offset-actual_offset)>size)
               read=size-bytes;
       rc = ext2fs_file_read(efile, buf+bytes,read , &bytes1);
       if (rc) {
          printf("READING FAILED %s \n",path);	
          return -EIO;
	
             }
  
       DEBUG_PRINTF("SUCCESS READING  %d \n",bytes1);
       if(bytes1==0)
             goto done;
       
       bytes=bytes1+bytes;
       offset=offset+bytes1;

       if(bytes>=size)
             {
               bytes=size;
               goto done;

             }           
        fflush(stdout); 
        goto back;
  
out1:
       DEBUG_PRINTF("ACTUAL OFFSET %d \n",actual_offset);
       DEBUG_PRINTF("Reading till  %d \n",read_offset); 

       rc = ext2fs_file_llseek(efile,actual_offset  , SEEK_SET, &pos);
       if (rc) {
		DEBUG_PRINTF("READING FAILED %s \n",path);
                return -EINVAL;
                }
       read=read_offset-actual_offset;
       if((bytes+read_offset-actual_offset)>size)
               read=size-bytes;
       rc = ext2fs_file_read(efile, buf+bytes,read , &bytes1);
       if (rc) {
          DEBUG_PRINTF("READING FAILED %s \n",path);	
          return -EIO;
	
             }
  
       DEBUG_PRINTF("SUCCESS READING %d \n",bytes1);
       if(bytes1==0)
             goto done;
       
       bytes=bytes1+bytes;
       offset=offset+bytes1;

       if(bytes>=size)
             {
               bytes=size;
               goto done;

             }           
        fflush(stdout); 
       goto back1;
       done:   
  	return bytes;
}


