/* UsiFe FileSystem
----------------
Copyright (2012): Rohan Sharma and Saurabh Kumar

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

CONTACT: 
Rohan Sharma : rickyrohan14@gmail.com
Saurabh Kumar : saurabh.catch@gmail.com
Smruti R. Sarangi : srsarangi@cse.iitd.ac.in
*/
#include "fuse_include.h" 
#include "do_xattr.h"

int do_setxattr(const char* path, const char* name, const char* value, size_t size, int flags)
{
 DEBUG_PRINTF("Entered setxattr %s \n",path);
 PWord_t  PValue; 

 errcode_t rc=-1;
 ext2_ino_t ino;
 ext2_filsys e2fs = (ext2_filsys)current_ext2fs();
 int val=*((int *)(value));
	
 pthread_mutex_lock(&read_inode_mutex);
  /* 
   * Read inode number of the file.
   */       
 rc = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, path, &ino);
 if (rc) {
	DEBUG_PRINTF("ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, %s, ino); failed %d\n", path,rc);
        pthread_mutex_unlock(&read_inode_mutex);  
        return -1;
	 }
 pthread_mutex_unlock(&read_inode_mutex);
 

 if(!strcmp(name,"Enc"))
 {
  JLI(PValue, PJLArray,ino);  
  val=val&ENC_MODE;
  DEBUG_PRINTF("Enc mode %d\n",val);
  if(PValue==PJERR)
    goto done;
  *PValue=(*PValue&(~ENC_MODE))|val; 
  rc=0;
 }
 else if(!strcmp(name,"Read"))
 {
  JLI(PValue, PJLArray,ino);  
  val=val&READ_MODE; 
  DEBUG_PRINTF("Read mode %d\n",val);
  if(PValue==PJERR)
    goto done;
  *PValue=(*PValue&(~READ_MODE))|val;
  rc=0;
 }
 
done:
 DEBUG_PRINTF("Leaving setxattr \n");

 return rc;
}

int do_getxattr(const char* path, const char* name, char* value, size_t size)
{
     __u64 pos;
    errcode_t rc;
    unsigned int bytes=0;
    ext2_ino_t ino;
    ext2_filsys e2fs = (ext2_filsys)current_ext2fs();
    struct ext2_inode inode;

    DEBUG_PRINTF("enter  getxattr %d \n",(int)size);
    DEBUG_PRINTF("path = %s\n\n", path);

    if(!isdigit(name[0]))
       goto exit;

    if(size==0)//if size is zero , return size of extended attributes
      {
       bytes=4096;
       goto exit;
      }

    pthread_mutex_lock(&write_mutex); 
    int rt = do_readinode(e2fs, path, &ino, &inode);
    if (rt) {
		DEBUG_PRINTF("do_read inode %s failed\n", path);
		pthread_mutex_unlock(&write_mutex); 
                return 0;
	}    
    pthread_mutex_unlock(&write_mutex); 

    

    if(inode.i_file_acl!=0)                         
         {
          DEBUG_PRINTF("Reading encrypted blocks\n");           

	  unsigned int enc_blk_no=atoi(name); //encrypted block no.
          short index=0;
          int i=0;         
          int start_blk=0;
          int off=0;
          char data[4096];

          start_blk=inode.i_file_acl;

	  for(i=0;i<enc_blk_no;i++)
	   {
	    pthread_mutex_lock(&mutex);
            io_channel_read_blk(e2fs->io,start_blk,1, data);
            pthread_mutex_unlock(&mutex);
 
            memcpy(&index,data,sizeof(short)); 
            if(index<=510)
               goto exit;
	    start_blk=*((int *)(data+4092)); 

	   }
	    pthread_mutex_lock(&mutex);
            io_channel_read_blk(e2fs->io,start_blk,1, data);
            pthread_mutex_unlock(&mutex);

           bytes= (4096) < size ? (4096) : size;

           memcpy(value,data,bytes);  
           DEBUG_PRINTF("Read enc blocks DONE %d\n",bytes); 
          }
	DEBUG_PRINTF("leave");
exit:
        fflush(stdout);
	return bytes;
}
