/* UsiFe FileSystem
----------------
Copyright (2012): Rohan Sharma and Saurabh Kumar

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

CONTACT: 
Rohan Sharma : rickyrohan14@gmail.com
Saurabh Kumar : saurabh.catch@gmail.com
Smruti R. Sarangi : srsarangi@cse.iitd.ac.in
*/
#include "usife.h"
/* 
 * Judy array for inode to flags(encryption mode , read mode ) mapping.
 */ 
#include <Judy.h>
/* 
 * FUSE related definitions
 */ 
#define FUSE_USE_VERSION 26
#define EXT2_FILE_SHARED_INODE 0x8000


/* 
 * FUSE headers
 */ 
#include <fuse.h>

#include <errno.h>
#include <fcntl.h>

#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>

/* Include openssl libraries to support encryption.
 * 
 */ 
#include <openssl/evp.h>
#include <openssl/aes.h>

/* Include gtk libraries to support GUI.
 * 
 */ 
#include <gtk/gtk.h>
#include <glib.h>
#include <pthread.h>

/* SSE2 instructions to support faster encryption.
 * 
 */ 
#include <xmmintrin.h>

/* 
 * DEBUG definitions
 */ 
#define DEBUG 1

#ifdef DEBUG
#define DEBUG_PRINTF printf
#else
#define DEBUG_PRINTF(format, args...) ((void)0)
#endif
 
/* 
 * Encryption and Reading mode definitions
 */ 
#define ENC_MODE     0x1
#define READ_MODE    0x6

#define ENC_ON       0x1
#define ENC_OFF      0x0

#define READ_MODE1   0x0
#define READ_MODE2   0x1
#define READ_MODE3   0x2


/* 
 * Definitions to have extended attributes functions or not. Since in our file system encrypted information is itself stored 
 * in extended attributes of inode(see do_write.c for detailed documentation) user is not expected to use these functions. Rather
 * usife library itself calls these functions to change encryption / read modes.
 */   
#define HAVE_SETXATTR


/* Variables defined in main file.
 * 
 */   

#ifndef USIFE_MAIN 
extern char key[];//encryption key 

extern pthread_mutex_t mutex,mutexr,change,mutex2,write_mutex,read_inode_mutex ;//mutex variables

extern pthread_mutex_t judy_mutex;//judy array mutex

extern ext2_filsys ext_fs;//filesystem

extern int mode;//file read mode

extern int e,e1;//encryption state

extern float e_p;//encryption density , only for debugging

extern Pvoid_t PJLArray;//JudyL array

extern int gtk_override;//just for debugging . It allows to use gtk user interface to change read & write(encryption) modes.
#endif


/* This structure defines encryption region of a file.
 * It is not being used anymore.Present here so that code compiles without any error.
 */ 
struct region{
    struct region *next;
    short   start;
    short   end; 
    int set; 
    };

/* 
 * This structure is used by fuse.
 */ 
struct extfs_data {
	unsigned char debug;
	unsigned char silent;
	unsigned char force;
	unsigned char readonly;
	time_t last_flush;
	char *mnt_point;
	char *options;
	char *device;
	char *volname;
	ext2_filsys e2fs;
};


/* 
 * Directory walk data, used while reading directory.
 */ 
struct dir_walk_data {
	char *buf;
	fuse_fill_dir_t filler;
};

/* 
 * From ext2fs library.
 */ 
//typedef struct ext2_file *ext2_file_t;


