/* UsiFe FileSystem
----------------
Copyright (2012): Rohan Sharma and Saurabh Kumar

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

CONTACT: 
Rohan Sharma : rickyrohan14@gmail.com
Saurabh Kumar : saurabh.catch@gmail.com
Smruti R. Sarangi : srsarangi@cse.iitd.ac.in
*/
#include "fuse_include.h"
#include "other.h"


ext2_file_t do_open2 (ext2_filsys e2fs, const char *path, int flags)
{
    ext2_ino_t ino;
    struct ext2_inode inode;
    int rt = do_readinode(e2fs,path, &ino, &inode);
   
    if(rt)
      { 
	DEBUG_PRINTF("failed to read inode %s\n",path);
        return NULL; 
      }
    ext2_file_t efile;

    rt = ext2fs_file_open2(e2fs, ino, &inode,
			(((flags & O_ACCMODE) != 0) ? EXT2_FILE_WRITE : 0) , 
			&efile);
    fflush(stdout);

    if(rt)
      { 
	DEBUG_PRINTF("failed to open file %s\n",path);
        return NULL; 
      }

    EXT2_CHECK_MAGIC(efile, EXT2_ET_MAGIC_EXT2_FILE);
    
    return (ext2_file_t)efile;
}


int do_readinode (ext2_filsys e2fs, const char *path, ext2_ino_t *ino, struct ext2_inode *inode)
{
	
        errcode_t rc;
	
        pthread_mutex_lock(&read_inode_mutex);
  	/* 
 	 * Read inode number of the file.
 	 */       
  	rc = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, path, ino);
	if (rc) {
		DEBUG_PRINTF("ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, %s, ino); failed %d\n", path,rc);
		
                pthread_mutex_unlock(&read_inode_mutex);  
                return -ENOENT;
	}
	
	/* 
	 * Read inode details.
 	 */ 
	rc = ext2fs_read_inode(e2fs, *ino, inode);
        pthread_mutex_unlock(&read_inode_mutex);       

       
    
        if (rc) {
		DEBUG_PRINTF("ext2fs_read_inode(e2fs, *ino, inode); failed");
		return -EIO;
	}
	return 0;
}


ext2_filsys current_ext2fs(void)
{
	struct fuse_context *mycontext=fuse_get_context();
	struct extfs_data *e2data=mycontext->private_data;
	return (ext2_filsys) e2data->e2fs;
}


void simple_encryption(const char *in,int start,int size,char *out) 
{

 int i=0,j=0;
 i=start-1;
 char k=0xff;

 DEBUG_PRINTF("Encryption %d %d\n",start,size); 

 
 int *a,*b;
 a=(int *)in;
 b=(int *)out;

 /* 
  * The for loop below takes xor of every int(4 bytes) with character 'k'(which is not a good way of encrypting data).
  * Since everytime an interger is being xored, data size should be multiple of 4.
  * You can even use better version of this for loop, by uncommenting the code below  which uses SSE2 instructions
  * to take xor of 16 bytes in one cycle.But again make sure data size is multiple of 16, otherwise glibc will report data
  * corruption error.
  */ 

 for(;i<start+size-1;)
     {
      *(b)=(*(a))^k;
      i=i+4;
      j=j+4;
      b=(int *)(out+j);
      a=(int *)(in +j);
     }



    /*  
    __m128 a,b;
    b=_mm_loadu_ps((float *)key);      
    i=0;
      do
      {
             a=_mm_loadu_ps((float *)(in+i));
             a=_mm_xor_ps(a,b);
             _mm_storeu_ps((float *)(out+i),a);
            i=i+16; 
      } while (i<4096);
    */
       
    fflush(stdout);

}


void start_encryption()
{
/**
 *Function for starting encryption region
 *Not implemented.
 */


}


void stop_encryption()
{
/**
 *Function for ending encryption region
 *Not implemented.
 */



}




int do_utimens (const char *path, const struct timespec tv[2])
{
    /*	int rt;
	errcode_t rc;
	ext2_ino_t ino;
	struct ext2_inode inode;
	ext2_filsys e2fs = current_ext2fs();

	rt = do_readinode(e2fs, path, &ino, &inode);
	if (rt) {
		printf("do_readvvvvvvvvvvvnnnnnnnnnode(%s, &ino, &vnode); failed\n", path);
		return rt;
	}
	
	
	inode.i_atime = tv[0].tv_sec;
	inode.i_mtime = tv[0].tv_sec;

	rt= ext2fs_write_inode (e2fs,ino, &inode);
        if (rc) {
		printf("inode time change ttttttttttt failed\n");
		return -EIO;
	}

	//printf("leave");
	return 0;*/
     return 0;
}

int do_chmod (const char *path, mode_t mode)

{

//no implemenation yet
/*
        int rt;
	int mask;
	time_t tm;
	errcode_t rc;
	ext2_ino_t ino;
	
	struct ext2_inode *inode;
	ext2_filsys e2fs = current_ext2fs();

	printf("enter");
	printf("path = %s 0%o", path, mode);

	rt = do_check(path);
	if (rt != 0) {
		printf("do_check(%s); failed", path);
		return rt;
	}

	rt = do_readvnode(e2fs, path, &ino, &vnode);
	if (rt) {
		printf("do_readvnode(%s, &ino, &vnode); failed", path);
		return rt;
	}
	inode = vnode2inode(vnode);

	tm = e2fs->now ? e2fs->now : time(NULL);
	mask = S_IRWXU | S_IRWXG | S_IRWXO | S_ISUID | S_ISGID | S_ISVTX;
	inode->i_mode = (inode->i_mode & ~mask) | (mode & mask);
	inode->i_ctime = tm;

	rc=vnode_put(vnode,1);
	if (rc) {
		printf("vnode_put(vnode,1); failed");
		return -EIO;
	}

	printf("leave");
	return 0;


*/
return 0;

}

int do_chown (const char *path, uid_t uid, gid_t gid)
{

//no implemenatation yet

        /*int rt;
	errcode_t rc;
	ext2_ino_t ino;
	struct ext2_vnode *vnode;
	struct ext2_inode *inode;
	ext2_filsys e2fs = current_ext2fs();

	printf("enter");
	printf("path = %s", path);
	
	rt = do_readvnode(e2fs, path, &ino, &vnode);
	if (rt) {
		printf("do_readvnode(%s, &ino, &vnode); failed", path);
		return rt;
	}
	inode = vnode2inode(vnode);
	
	if (uid != -1)
		inode->i_gid = gid;
	if (gid != -1)
		inode->i_uid = uid;

	rc=vnode_put(vnode,1);
	if (rc) {
		printf("vnode_put(vnode,1); failed");
		return -EIO;
	}

	printf("leave");
	return 0;*/






 return 0;

} 

