/* UsiFe FileSystem
----------------
Copyright (2012): Rohan Sharma and Saurabh Kumar

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

CONTACT: 
Rohan Sharma : rickyrohan14@gmail.com
Saurabh Kumar : saurabh.catch@gmail.com
Smruti R. Sarangi : srsarangi@cse.iitd.ac.in
*/
/**
 * Copyright (c) 2011-2020 Rohan Sharma <rickyrohan14@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the fuse-ext2
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define USIFE_MAIN

/* 
 * All needed includes and structures
 */ 
#include "fuse_include.h"

/* 
 * Includes initialization function
 */ 
#include "do_init.h"

/* 
 * Block allocation 
 */ 
#include "alloc.h"

/* 
 * Helper functions
 */ 
#include "other.h"

/* 
 * Directory read functions
 */ 
#include "do_readdir.h"

/* 
 * Read File attributes
 */ 
#include "do_getattr.h"

/* 
 * File open function
 */ 
#include "do_open.h"


/* 
 * AES encryption
 */ 
#include "aes.h"

/* 
 * GTK interface
 */ 
#include "gui.h"

/* 
 * Create files and directories.
 */ 
#include "do_create.h"

/* 
 * Delete files and directories.
 */ 
#include "do_delete.h"

/* 
 * File system statistics.
 */ 
#include "do_statfs.h"

/* 
 * Closes filesystem
 */ 
#include "do_destroy.h"

/* 
 * Read file
 */ 
#include "do_read.h"

/* 
 * Releasing file handles
 */ 
#include "do_release.h"

/* 
 * Write data to a file.
 */ 
#include "do_write.h"

/* 
 * Functions to change encryption and read modes.
 */ 
#include "do_xattr.h"

int added=0;

char key[4096];//encryption key 

pthread_mutex_t mutex,mutexr,change,mutex2,write_mutex,read_inode_mutex ;//mutex variables
pthread_mutex_t judy_mutex;
ext2_filsys ext_fs;//filesystem

int mode=2;//file read mode

int e=0,e1=0;//encryption state

float e_p=0.5;//encryption density , only for debugging

Pvoid_t PJLArray = (Pvoid_t) NULL;//initialize JudyL array

int gtk_override=0;//just for debugging . It allows to use gtk user interface to change read & write(encryption) modes.

static struct fuse_operations ext2_oper = {
      .init      = do_init,
      .getattr   = do_getattr,
      .readdir   = do_readdir,
      .open      = do_open,
      .read      = do_read,
      .write     = do_write,
      .utimens   = do_utimens,    
      .destroy	 = do_destroy,
      .chmod     = NULL,
      .chown     = NULL,
      .truncate  = do_truncate,   
      .release   = do_release,
      .mkdir     = do_mkdir,
      .create    = do_create,
      .unlink    = NULL,
      .rmdir     = do_rmdir,   
      .statfs    = do_statfs,
      .rename    = NULL,
      #ifdef HAVE_SETXATTR
      .setxattr  = do_setxattr,
      .getxattr  = do_getxattr,
      #endif
  };
  
int main(int argc, char *argv[])
{
    
    struct extfs_data opt;  

  /* user interface code
   * 
   */
    GtkWidget *window = NULL;
    GtkWidget *box1;
    GtkWidget *box2;
    GtkWidget *button;
    GtkWidget *label; 
    GtkWidget *separator;
    GSList *group;

    GtkWidget  *spinner;
    GtkAdjustment *spinner_adj;

    
    int a=1,b=2,c=3;

    pthread_t p;
  
    gtk_init (&argc, &argv);    
      
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  
    g_signal_connect (window, "delete-event",
		      G_CALLBACK (close_application),
                      NULL);

    gtk_window_set_title (GTK_WINDOW (window), "File Encryption");
    gtk_container_set_border_width (GTK_CONTAINER (window), 0);

    box1 = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (window), box1);
    gtk_widget_show (box1);

    box2 = gtk_vbox_new (FALSE, 10);
    gtk_container_set_border_width (GTK_CONTAINER (box2), 10);
    gtk_box_pack_start (GTK_BOX (box1), box2, TRUE, TRUE, 0);
    gtk_widget_show (box2);

    button = gtk_radio_button_new_with_label (NULL, "button1");
    g_signal_connect (button, "clicked",
		      G_CALLBACK (switch_mode), (gpointer)&a);
    gtk_box_pack_start (GTK_BOX (box2), button, TRUE, TRUE, 0);
    gtk_widget_show (button);

    group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (button));
    button = gtk_radio_button_new_with_label (group, "button2");
    g_signal_connect (button, "clicked",
		      G_CALLBACK (switch_mode), (gpointer)&b);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
    gtk_box_pack_start (GTK_BOX (box2), button, TRUE, TRUE, 0);
    gtk_widget_show (button);

    button = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (button),
	                                                  "button3");
    g_signal_connect (button, "clicked",
		      G_CALLBACK (switch_mode), (gpointer)&c);
    gtk_box_pack_start (GTK_BOX (box2), button, TRUE, TRUE, 0);
    gtk_widget_show (button);

    separator = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (box1), separator, FALSE, TRUE, 0);
    gtk_widget_show (separator);

    box2 = gtk_vbox_new (FALSE, 10);
    gtk_container_set_border_width (GTK_CONTAINER (box2), 10);
    gtk_box_pack_start (GTK_BOX (box1), box2, FALSE, TRUE, 0);
    gtk_widget_show (box2);

    label=gtk_label_new("Encryption OFF");
    gtk_box_pack_start (GTK_BOX (box2), label, TRUE, TRUE, 0);
    gtk_widget_show (label);

    
    button = gtk_button_new_with_label ("ON/Off Encryption");
    g_signal_connect (button, "clicked",
		      G_CALLBACK (switch_enc), (gpointer)label);
    gtk_box_pack_start (GTK_BOX (box2), button, TRUE, TRUE, 0);
    gtk_widget_set_can_default (button, TRUE);
    gtk_widget_grab_default (button);
    gtk_widget_show (button);

    button = gtk_button_new_with_label ("Dump");
    g_signal_connect (button, "clicked",
		      G_CALLBACK (dump_acl), NULL);
    gtk_box_pack_start (GTK_BOX (box2), button, TRUE, TRUE, 0);
    gtk_widget_set_can_default (button, TRUE);
    gtk_widget_grab_default (button);
    gtk_widget_show (button);
    

    spinner_adj = (GtkAdjustment *) gtk_adjustment_new (0.50, 0.00, 1.00, 0.05, 0.05, 0.05);
    spinner = gtk_spin_button_new (spinner_adj, 0.05, 2);
    gtk_box_pack_start (GTK_BOX (box2), spinner, TRUE, TRUE, 0);
    gtk_widget_show (spinner);

    button = gtk_button_new_with_label ("SET");
    g_signal_connect (button, "clicked",
		      G_CALLBACK (set_enc), (gpointer)spinner);
    gtk_box_pack_start (GTK_BOX (box2), button, TRUE, TRUE, 0);
    gtk_widget_set_can_default (button, TRUE);
    gtk_widget_grab_default (button);
    gtk_widget_show (button);


    gtk_widget_show (window);
     

    
    /** user interface code
     * 
     */
    

    return fuse_main(argc, argv, &ext2_oper, &opt);
}

